/*
 * $Header: /home/puhuri/Tutkimus/src/timer/RCS/printout.c,v 1.4 1997/02/27 13:18:18 puhuri Exp puhuri $
 *
 * $Log: printout.c,v $
 * Revision 1.4  1997/02/27 13:18:18  puhuri
 * Listty ajan laskenta oikein.
 *
 * Revision 1.3  1997/02/19 12:47:47  puhuri
 * Ekojen mittausten versio.
 *
 * Revision 1.2  1996/11/22 08:26:36  puhuri
 * Muutettu kyttmn conf-rakennetta ja yhteist tulostustapa
 *
 * Revision 1.1  1996/11/14 10:33:13  puhuri
 * Initial revision
 *
 *
 *
 */
#include <sys/time.h>
#include <stdio.h>
#include <unistd.h>
#include "sched_tst.h"

static FILE *po_open (const char *ofile_fmt, char mytype, int mynro);

static struct timeval difftimei (struct timeval t1, struct timeval t2)
{
  struct timeval tmp;

  tmp.tv_sec = t1.tv_sec - t2.tv_sec;
  tmp.tv_usec = t1.tv_usec - t2.tv_usec;
  if (tmp.tv_usec < 0) {
    tmp.tv_sec--;
    tmp.tv_usec += i10E6;
  }
  if (tmp.tv_usec > i10E6) {
    tmp.tv_sec++;
    tmp.tv_usec -= i10E6;
  }
  return tmp;
}
/* 
 * Tulostetaan aika, timeval-structuuri
 */

void printout_time_tval (const struct conf_info *conf, 
			 const struct timeval *tvp,
			 int loops, int mynro, char mytype)
{
  int i;
  FILE *fout;
  struct timeval t1, t2;

  t1 = add_time (conf->start_time, conf->drift);
  fprintf (stderr, "%d:%c %ld.%06ld\n",
	   getpid(), mytype, (long)t1.tv_sec, (long)t1.tv_usec);

  if ((fout = po_open (conf->ofile_fmt, mytype, mynro)) == NULL) {
    exit (ERR_FILEOPEN);
  }
  
  for (i = 0; loops > i; i++) {	/* tulostetaan kaikki */
    if (fprintf (fout, "%ld.%06ld\n", 
		 (long)tvp[i].tv_sec, (long)tvp[i].tv_usec) < 1) {
      PERROR ("fprintf: ", ERR_FILEWRITE);
    }
  }

  if (fclose (fout)) {
      PERROR ("fclose: ", ERR_FILECLOSE);
  }

  t1 = difftimei (tvp[0], conf->start_time);
  t2 = difftimei (tvp[loops-1], conf->start_time);

  fprintf (stderr, "%d:%c %d written (%ld.%06ld -- %ld.%06ld)\n",
	   getpid(), mytype, loops, (long)t1.tv_sec, (long)t1.tv_usec,
	    (long)t2.tv_sec, (long)t2.tv_usec);
}

/* 
 * Tulostetaan aika, log-struktuuri
 */

void printout_time_log (const struct conf_info *conf, 
			const struct log_entry *log,
			int loops, int mynro, char mytype)
{
  int i;
  FILE *fout;
  struct timeval t1, t2;

  t1 = add_time (conf->start_time, conf->drift);

  fprintf (stderr, "%d:%c %ld.%06ld\n",
	   getpid(), mytype, (long)t1.tv_sec, (long)t1.tv_usec);

  if ((fout = po_open (conf->ofile_fmt, mytype, mynro)) == NULL) {
    exit (ERR_FILEOPEN);
  }

  for (i = 0; loops > i; i++) {	/* tulostetaan kaikki */
    t1 = add_time (log[i].timeval, conf->drift);

    if (fprintf (fout, "%ld.%06ld\t%d\n", 
		 (long)t1.tv_sec, (long)t1.tv_usec,
		 log[i].read_bytes)< 1) {
      PERROR ("fprintf: ", ERR_FILEWRITE);
    }
  }
  
  if (fclose (fout)) {
      PERROR ("fclose: ", ERR_FILECLOSE);
  }

  t1 = difftimei (log[0].timeval, conf->start_time);
  t2 = difftimei (log[loops-1].timeval, conf->start_time);

  fprintf (stderr, "%d:%c %d written (%ld.%06ld -- %ld.%06ld)\n",
	   getpid(), mytype, loops, (long)t1.tv_sec, (long)t1.tv_usec,
	    (long)t2.tv_sec, (long)t2.tv_usec);
}

/*
 * Luodaan tiedoston nimi ja avataan se
 */

static FILE *po_open (const char *ofile_fmt, char mytype, int mynro)
{
  char ofile[BUF_SIZE];
  FILE *fout;

  sprintf(ofile, ofile_fmt, mytype, mynro);	/* luodaan tiedostonimi */

  DPRINTF ("pid %d to file '%s' fmt='%s', nro=%d, type=%c", 
	   getpid(), ofile, ofile_fmt, mynro, mytype);
  if ((fout = fopen(ofile, "w")) == NULL) {
    PERROR ("fopen: ", ERR_NOEXIT); /* tulostetaan virheilmoitus */
    fprintf (stderr, "file = '%s'\n", ofile);
    return NULL;
  }
  return fout;
}
