/*
 * $Header: /home/puhuri/Tutkimus/src/timer/RCS/sink.c,v 1.5 1997/02/27 13:29:48 puhuri Exp puhuri $
 *
 * $Log: sink.c,v $
 * Revision 1.5  1997/02/27 13:29:48  puhuri
 * Luetaan niin kauen kuin saadaan.  (Vltetn lhettvn SIGPIPE).
 *
 * Revision 1.4  1997/02/19 11:53:23  puhuri
 * Ensimmisten mittausten versio.
 *
 * Revision 1.3  1996/11/22 08:30:03  puhuri
 * Muutettu uuteen rakenteeseen, mm lyt itse tpselins.
 *
 * Revision 1.2  1996/11/13 12:01:49  puhuri
 * Changed to work.
 *
 * Revision 1.1  1996/11/05 17:47:55  puhuri
 * Initial revision
 *
 *
 *
 */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include "sched_tst.h"

#ifdef RCS_INFO
static char rcs_id[] = "$Id: sink.c,v 1.5 1997/02/27 13:29:48 puhuri Exp puhuri $";
#endif

void estimate_start_time (struct conf_info *conf,
			  struct time_message *tmsg);

struct time_message *sync_time_client (void);

int sink (struct conf_info *conf)
{
  struct log_entry *log;
  char *read_buf, pbuf[PORT_SIZE + 1 + 1];
  int inb, sa_size, i, mynro = 0;
  int fd, rfd;			/* listen fd, read fd */
  struct sockaddr_in my_addr, source_addr;
  struct time_message *tmsg;
  int read_size = conf->read_size;
  int port;
  int wtimes = conf->write_times;
  int before_reads = 0;
  int before_bytes = 0;

  if ((read_buf = malloc(read_size)) == NULL) {
    PERROR ("malloc: ", ERR_NO_MEMORY);
  }

  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    PERROR("socket:", ERR_SOCKET);
  }

  my_addr.sin_family = AF_INET;
  my_addr.sin_addr.s_addr = htonl(INADDR_ANY);

  for (port = IPPORT_RESERVED; port < 0xffff; port++) {
    my_addr.sin_port = htons (port);
    if (!(bind(fd,  (struct sockaddr *)&my_addr, sizeof(my_addr)) 
	  < 0)) {
      break;
    }
    if (port == 0xffff) {
      PERROR ("bind", ERR_BIND);
    }
  }

  if ( listen (fd, IN_QUEUE) ) {
    PERROR ("listen: ", ERR_LISTEN);
  }

  DPRINTF ("Reader: my listening port %d (%d)", port, getpid());
  sprintf (pbuf, "%0" STROF(PORT_SIZE) "d\n", port);
  write (STDOUT_FILENO, pbuf, PORT_SIZE + 1); /* kerrotaan portti */

  tmsg = sync_time_client ();
  estimate_start_time (conf, tmsg);

  while (1) {
    fd_set readfds;
    int rtr;			/* ready to read */
    pid_t pid;

    FD_ZERO (&readfds);
    FD_SET (fd, &readfds);
    FD_SET (STDIN_FILENO, &readfds);

    for (rtr = select (fd + 1, &readfds, NULL, NULL, NULL);
	 rtr > 0; rtr --) {
      if (FD_ISSET (fd, &readfds)) {
	sa_size = sizeof(source_addr);
	if ((rfd = accept (fd, (struct sockaddr *)&source_addr, &sa_size)) 
	    < 0) {
	  PERROR ("accept: ", ERR_ACCEPT);
	}
    
	if ((pid = fork()) == 0) {
	  if ((log = get_dirty_mem (wtimes * sizeof(struct log_entry))) 
	      == NULL) {
	    PERROR ("malloc: ", ERR_NO_MEMORY);
	  }
	  
	  i = 0;
	  while ((inb = read (rfd, read_buf, read_size)) > 0) {
	    if (i >= wtimes)	/* valvotaan, ettei ylity */
	      continue;		/* luetaan niin kauan kuin pystytn */

	    if (gettimeofday (&(log[i].timeval), NULL) < 0) {
	      PERROR ("gettimeofday: ", ERR_GETTIMEOFDAY);
	    }
	    log[i].read_bytes = inb;

	    if (timercmp(&(log[i].timeval), &(conf->go_time), <)) {
	      before_bytes += inb;	/* paljonkos lhetetn ennen */
	      before_reads++;	
	      continue;
	    }
				/* ei tarvetta mitata loppumisaikaa,
				 * luetaan niin paljon kuin on
				 * tullakseen
				 */
	    ++i;		/* kasvatetaan laskuria */
	  }
	  close (rfd);

	  if (conf->usleep_end > 0) { /* maybe sleep some... */
#ifdef _HPUX_SOURCE		/* HPUX: ei usleep */
	    sleep (conf->usleep_end / i10E6 ? : 1);
#else
	    usleep (conf->usleep_end);
#endif
	  }
	  printout_time_log (conf, log, i - 1, mynro, ID_CHAR_RECV);
	  fprintf (stderr, 
		   "%d:%c mynro = %d port = %d: before writes = %d (%d bytes)\n",
		   getpid(), ID_CHAR_RECV, mynro, ntohs (source_addr.sin_port),
		   before_reads, before_bytes);

	  exit (OK);
	} else {			/* parent */
	  close (rfd);
	  add_child_pid (conf, pid);
	  mynro ++;
	}
      } else if (FD_ISSET(STDIN_FILENO, &readfds)) {
	char buf[SBUF_SIZE];

	if (read (STDIN_FILENO, buf, 1) < 0) {
	  PERROR ("read", ERR_PIPEREAD);
	}
	if (*buf == 'x') {
	  for (;;) {
	    int child, child_exit_status;

	    child = wait (&child_exit_status);
	    if (decode_wait (child, child_exit_status) == OK) {
	      int i;
	      for (i = 0; i < UDP_COUNT * 2; i++)
		fprintf (stderr, "tmsg[%d] := %ld.%06ld\n",
			 i, (long)tmsg[i].time.tv_sec,
			 (long)tmsg[i].time.tv_usec);
	      exit(OK);
	    }
	  }
	}
      } else {
	PERROR ("select", ERR_SELECT);
      }
    }
  }
  exit (OK);
}

struct time_message *sync_time_client (void)
{
  int udp;			/* udp datagram fd */
  int udp_port;
  struct sockaddr_in udp_my_addr, udp_other_addr;
  int sockaddr_len;
  char pbuf[PORT_SIZE + 1 + 1];
  struct time_message *tvbuf;
  int i;			/* laskuri */

  if ((udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    PERROR ("socket (dgram):", ERR_SOCKET);
  }

  udp_my_addr.sin_family = AF_INET;
  udp_my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  /*
   * Etsitn vapaa UDP-portti
   */
  for (udp_port = IPPORT_RESERVED; udp_port < 0xffff; udp_port++) {
      udp_my_addr.sin_port = htons (udp_port);
      if (!(bind(udp,  (struct sockaddr *)&udp_my_addr, sizeof(udp_my_addr)) 
	    < 0)) {
	break;
      }
      if (udp_port == 0xffff) {
	PERROR ("bind (udp)", ERR_BIND);
      }
  }
  sprintf (pbuf, "%0" STROF(PORT_SIZE) "d", udp_port);
  write (STDOUT_FILENO, pbuf, PORT_SIZE + 1); /* kerrotaan portti */

  if ((tvbuf = malloc (sizeof (struct time_message) * UDP_COUNT*2)) == NULL) {
    PERROR ("malloc:", ERR_NO_MEMORY);
  }

  sockaddr_len = sizeof (struct sockaddr_in);
  if (recvfrom (udp, &tvbuf[0], sizeof(struct time_message), 0,
		(struct sockaddr *)&udp_other_addr, &sockaddr_len) < 0) {
    PERROR ("recvfrom: ", ERR_RECVFROM);
  }

  for (i = 0; i < UDP_COUNT - 1; i++) {
    fd_set readfds;
    struct timeval timeout;

    tvbuf[i*2+1].number = i + 1;
    if (gettimeofday (&tvbuf[i*2+1].time, NULL) < 0) {
      PERROR ("gettimeofday: ", ERR_GETTIMEOFDAY);
    }
    if (sendto (udp, &tvbuf[i*2+1], sizeof (struct time_message), 0,
		(struct sockaddr *)&udp_other_addr, sockaddr_len) < 0) {
      PERROR ("sendto: ", ERR_SENDTO);
    }
    timeout.tv_sec = UDP_SELECT_TIMEOUT_SEC;
    timeout.tv_usec = UDP_SELECT_TIMEOUT_USEC;
    FD_ZERO (&readfds);
    FD_SET (udp, &readfds);
    if (select (udp + 1, &readfds, NULL, NULL, &timeout) == 0) {
      DPRINTF ("Message lost (missing number %d)", i);
      i = -1;			/* viesti kadonnut, uudestaan */
      continue;
    } else {
      if (read (udp, &tvbuf[i*2+2], sizeof (struct time_message)) < 0) {
	PERROR ("read: ", ERR_READ);
      }
    }
  }
  return tvbuf;
}


#if 0
#define LITTLE_ENDIAN   1234    /* least-significant byte first (vax) */
#define BIG_ENDIAN      4321    /* most-significant byte first (IBM, net) */
#define PDP_ENDIAN      3412    /* LSB first in word, MSW first in long (pdp) */
#endif

/*
 * sama tavujrjestyt
 */
int xe2xe (int x) {
  return x;
}
/*
 * LITTLE_ENDIAN -> BIG_ENDIAN
 * BIG_ENDIAN -> LITTLE_ENDIAN
 */
int le2be (int x) {
  int t;

  t = (x & 0xff) << 24;
  t |= ((x >> 8) & 0xff) << 16;
  t |= ((x >> 16) & 0xff) << 8;
  t |= ((x >> 24) & 0xff);

  return t;
}
/*
 * LITTLE_ENDIAN -> PDP_ENDIAN
 * PDP_ENDIAN -> LITTLE_ENDIAN
 */
int le2pe (int x) {
  int t;

  t = (x & 0xff) << 16;
  t |= ((x >> 8) & 0xff) << 24;
  t |= ((x >> 16) & 0xff);
  t |= ((x >> 24) & 0xff) << 8;

  return t;
}
/*
 * BIG_ENDIAN -> PDP_ENDIAN
 * PDP_ENDIAN -> BIG_ENDIAN
 */
int be2pe (int x) {
  return le2pe(le2be(x));
}

typedef int (*endian_convert_t)(int);
endian_convert_t ec_table[] = {
  xe2xe, le2be, le2pe, be2pe, NULL
};

long t_av12_3 (struct timeval t1, struct timeval t2, struct timeval t3) 
{
  long long v1s, v2s, v3s;
  long long v1u, v2u, v3u;

  v1s = (long long)t1.tv_sec + (long long)t2.tv_sec;
  v2s = (long long)t3.tv_sec * 2;
  v3s = (v1s - v2s) * i10E6 / 2;

  v1u = (long long)t1.tv_usec + (long long)t2.tv_usec;
  v2u = (long long)t3.tv_usec * 2;
  v3u = (v1u - v2u) / 2;

  return (long)(v3s + v3u);
}

void estimate_start_time (struct conf_info *conf,
			  struct time_message *tmsg)
{
  int i, j;
  int diffs[UDP_COUNT];
  int sort[UDP_COUNT];
  int ave_err;
  endian_convert_t convert = NULL;

  for (i = 0; ec_table[i] != NULL; i++) {
    if ((*ec_table[i])(tmsg[0].number) == ENDIAN_MAGIC_NUMBER) {
      convert = ec_table[i];
      break;
    }
  }
  if (ec_table[i] == NULL) {
    fprintf (stderr, "Cant convert endians: 0x%08x != 0x%08x\n",
	     ENDIAN_MAGIC_NUMBER, tmsg[0].number);
    exit (ERR_BAD_ENDIAN);
  }

  for (i = 0; i < UDP_COUNT * 2; i++) {
    tmsg[i].number = (*convert)(tmsg[i].number);
    tmsg[i].time.tv_sec = (*convert)(tmsg[i].time.tv_sec);
    tmsg[i].time.tv_usec = (*convert)(tmsg[i].time.tv_usec);
  }

  for (i = 0; i < UDP_COUNT - 2; i += 2) {
    diffs[i] = t_av12_3 (tmsg[i+3].time, tmsg[i+1].time, tmsg[i+2].time);
    /* x = (t3 + t1)/2 - t2 */

    if (i < UDP_COUNT - 3 )	/* x = t3 - (t2+t4)/2  = - ((t2+t4)/2 - t3) */
      diffs[i+1] = - t_av12_3 (tmsg[i+2].time, tmsg[i+4].time, 
			       tmsg[i+3].time);
  }
  /*
   * Lajitellaan arvot
   */
  for (i = 0; i < UDP_COUNT - 2; i++) {
    for (j = i; (j > 0) && (sort[j] > diffs[i]); j--)
      sort[j-1] = sort[j];
    sort[j] = diffs[i];
    DPRINTF ("diff[%d] = %d", j, diffs[i]);
  }
  for (i = (UDP_COUNT - 2) / 4, j = 0, ave_err = 0;
       i < (UDP_COUNT - 2) - ((UDP_COUNT - 2) / 4); i++) {
    ave_err += sort[i];
    j++;
  }

  ave_err = ave_err / j;	/* keskimrinen ero */
  DPRINTF("Average error seems to be %d.%06d seconds",
	  (ave_err >= 0)? ave_err / i10E6: -(abs(ave_err)/i10E6-1),
	  (ave_err >= 0)? ave_err % i10E6: i10E6 - (abs(ave_err) % i10E6));
  DPRINTF("this: %d.%06d", conf->start_time.tv_sec, conf->start_time.tv_usec);
  /* tassa_koneessa = toisessa_koneessa + erotus */
  DPRINTF("that: %d.%06d", tmsg[0].time.tv_sec, tmsg[0].time.tv_usec);

  conf->start_time = add_time (tmsg[0].time, ave_err);
  DPRINTF("corr: %d.%06d", conf->start_time.tv_sec, conf->start_time.tv_usec);

  conf->drift = ave_err;

  conf->go_time = add_time (conf->start_time, conf->usleep_start);
  DPRINTF("go:   %d.%06d", conf->go_time.tv_sec, conf->go_time.tv_usec);
  DPRINTF("now:  %d.%06d", tmsg[UDP_COUNT-1].time.tv_sec, 
	  tmsg[UDP_COUNT-1].time.tv_usec);

  return;
}

/*
 * Server end.
 */

void sync_time_server (struct conf_info *conf)
{
  int udp;			/* udp datagram fd */
  int udp_port;
  struct sockaddr_in udp_my_addr, udp_other_addr;
  int sockaddr_len;
  char pbuf[PORT_SIZE + 1 + 1];
  struct time_message *tvbuf;
  struct hostent *he;
  int i;

  if ((udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    PERROR ("socket (dgram):", ERR_SOCKET);
  }

  udp_my_addr.sin_family = AF_INET;
  udp_my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  /*
   * Etsitn vapaa UDP-portti
   */
  for (udp_port = IPPORT_RESERVED; udp_port < 0xffff; udp_port++) {
      udp_my_addr.sin_port = htons (udp_port);
      if (!(bind(udp,  (struct sockaddr *)&udp_my_addr, sizeof(udp_my_addr)) 
	    < 0)) {
	break;
      }
      if (udp_port == 0xffff) {
	PERROR ("bind (udp)", ERR_BIND);
      }
  }

  DPRINTF ("Local UDP port: %d (%d)", udp_port, getpid());

  read (conf->fd_from_remote, pbuf, PORT_SIZE + 1); 
  /* otetaan selville portti */
  DPRINTF ("Remote UDP port: %d (%d)", atoi(pbuf), getpid());

  if ((tvbuf = malloc (sizeof (struct time_message) * UDP_COUNT*2)) == NULL) {
    PERROR ("malloc:", ERR_NO_MEMORY);
  }

  if ((he = gethostbyname ( conf->host )) == NULL) {
    PERROR ("gethostbyname: ", ERR_GETHOSTBYNAME);
  }
  udp_other_addr.sin_family = AF_INET;
  udp_other_addr.sin_addr = *((struct in_addr *)he->h_addr);
  udp_other_addr.sin_port = htons (atoi(pbuf));

  tvbuf[0].number = ENDIAN_MAGIC_NUMBER;
  tvbuf[0].time = conf->start_time;
  sockaddr_len = sizeof (struct sockaddr_in);
  if (sendto (udp, &tvbuf[0], sizeof(struct time_message), 0,
		(struct sockaddr *)&udp_other_addr, sockaddr_len) < 0) {
    PERROR ("sendto: ", ERR_SENDTO);
  }

  for (i = 0; i < UDP_COUNT - 1; i++) {
    while (read (udp, &tvbuf[i*2+2], sizeof (struct time_message)) < 0) {
      if (errno != EINTR)	/* jos saadaan signaali */
	PERROR ("read: ", ERR_READ);
    }

    if (tvbuf[i*2+2].number == 0) { /* jos viesti hukkui, aloitetaan */
      i = -1;			/* alusta */
      continue;
    }
    tvbuf[i*2+1].number = i + 1;
    if (gettimeofday (&tvbuf[i*2+1].time, NULL) < 0) {
      PERROR ("gettimeofday: ", ERR_GETTIMEOFDAY);
    }

    if (sendto (udp, &tvbuf[i*2+1], sizeof (struct time_message), 0,
		(struct sockaddr *)&udp_other_addr, sockaddr_len) < 0) {
      PERROR ("sendto: ", ERR_SENDTO);
    }
  }
}
