/*
 * $Id: test_traffic.c,v 1.5 1997/05/14 12:48:46 puhuri Exp puhuri $
 * $Log: test_traffic.c,v $
 * Revision 1.5  1997/05/14 12:48:46  puhuri
 * Toimii, uudempi versio.
 *
 * Revision 1.4  1997/05/14 07:03:57  puhuri
 * Ekojen mittausten versio.
 *
 * Revision 1.3  1996/12/02 09:19:57  puhuri
 * Linja- ja kehysnopeus vakioitu.
 *
 * Revision 1.2  1996/11/29 13:35:31  puhuri
 * Toimiva versio.
 *
 * Revision 1.1  1996/11/29 10:56:42  puhuri
 * Initial revision
 *
 * Revision 1.1  1996/11/27 10:42:47  puhuri
 * Initial revision
 *
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include "sched_tst.h"

int not_tried (tr_param *tr, int d_inv_sbr,  int d_mbs);
cell_log read_file (char *fname, int bytes, int inv_lcr, int inv_fcr,
		    int count);
static inline void usage() {printf( "Error in args\n"); exit(100);}

#define MAX_TRIES 200
#define CELLS_PER_FRAME 31
#define SCALE 1000


int main (int argc, char **argv)
{
  int bytes = 1500;
  int inv_lcr;			/* linjanopeus (solujen vlit) */
  int inv_fcr;			/* nopeus, jolla kehyksen solut lhetetn */
  int inv_clp = -1;
  double lbr = 155520000.0;
  double fbr = 1.0e7;
  double pbr = 1.0e7;
  double sbr;
  int test;
  int try;			/* kokeilukerrat */
  int d_inv_sbr;
  int d_mbs;
  char option;
  char *infile = NULL;

  tr_param tr, ok_tr;
  cell_log cl;

#define OPTSTRING "b:l:f:p:F:"
  while ((option = getopt (argc, argv, OPTSTRING)) != -1) {
    switch (option) {
      case 'b':
	bytes = atoi(optarg);
	break;
    case 'l':
      lbr =  atof(optarg);
      break;
    case 'f':
      fbr =  atof(optarg);
      break;
    case 'p':
      pbr = atof(optarg);
      break;
    case 'F':
      infile = optarg;
      break;
    default: usage();
    }
  }

  if (lbr == 0.0 || fbr == 0.0 || pbr == 0.0) {
    printf ("Something strange zeros\n");
    exit(12);
  }
  if (infile == NULL) {
    printf ("Something strange files\n");
    exit(12);
  }

  inv_lcr = (int)(1.0 / (lbr / BITS_PER_CELL) * TUNITS);
  inv_fcr = (int)(1.0 / (fbr / BITS_PER_CELL) * TUNITS);
  tr.inv_pcr = (int)(1.0 / (pbr / BITS_PER_CELL) * TUNITS);
  tr.cdvt = inv_fcr * CELLS_PER_FRAME;

  cl = read_file (infile, bytes, inv_lcr, inv_fcr, 10000);
  for (test = optind; test < argc; test += 4){
    sbr = atof(argv[test]);
    d_inv_sbr = atoi(argv[test+1]);
    tr.mbs = atoi(argv[test+2]);
    d_mbs = atoi(argv[test+3]);

    tr.inv_scr = (int)(1.0 / (sbr / BITS_PER_CELL) * TUNITS);
#ifdef DEBUG
    printf ("Line rate = %g b/s = %g c/s = %d " TUNIT "/c\n",
	    lbr, (lbr / BITS_PER_CELL), inv_lcr);
    printf ("Frame cell rate = %g b/s = %g c/s = %d " TUNIT "/c\n",
	    fbr, (fbr / BITS_PER_CELL), inv_fcr);
    printf ("Peak rate = %g b/s = %g c/s = %d " TUNIT "/c\n",
	    pbr, (pbr / BITS_PER_CELL), tr.inv_pcr);
    printf ("Suistanable rate = %g b/s = %g c/s = %d " TUNIT "/c\n",
	    sbr, (sbr / BITS_PER_CELL), tr.inv_scr);
    printf ("Max Burst Size = %d bits = %d cells\n", tr.mbs * BITS_PER_CELL,
	    tr.mbs);
    printf ("Cell Delay Variation Tolerance = %d " TUNIT "\n", 
	    tr.cdvt);
    printf ("%d cells time = %f (%f..%f)\n", cl.max + 1,
	    (double)(cl.table[cl.max] - cl.table[0]) / TUNITS,
	    (double)(cl.table[0]) / TUNITS,
	    (double)(cl.table[cl.max]) / TUNITS);
#endif
    printf (
#ifdef DEBUG
	    "Suistanable rate = %g b/s = %g c/s = %g " TUNIT "/c\n",
#else
	    "\t%g",
#endif
	    (cl.max + 1) * BITS_PER_CELL / 
	    ((double)(cl.table[cl.max] - cl.table[0]) / TUNITS)
#ifdef DEBUG
	    ,
	    (cl.max + 1) / ((double)(cl.table[cl.max] - cl.table[0]) / TUNITS),
	    ((double)(cl.table[cl.max] - cl.table[0])) / (cl.max + 1)
#endif
	    );

    ok_tr.inv_pcr = -1;
    not_tried (NULL, d_inv_sbr, d_mbs);
    for (try = 0; try < MAX_TRIES && 
	   not_tried (&tr, d_inv_sbr, d_mbs); try++) {
      inv_clp = not_withn_contract (cl, tr);
      if (inv_clp) {		/* ei tyt sopimusta: scr-, mbs+ */
	if (d_inv_sbr != 0)
	  tr.inv_scr -=  d_inv_sbr;
	if (d_mbs != 0)
	  tr.mbs += d_mbs;
      } else {			/* meni lpi, puolitetaan askel */
	ok_tr = tr;
	if (d_inv_sbr) {
	  d_inv_sbr >>= 1;
	  if (!d_inv_sbr)
	    d_inv_sbr = 1;
	  tr.inv_scr += d_inv_sbr;
	}
	if (d_mbs) {
	  d_mbs >>= 1;
	  if (!d_mbs)
	    d_mbs = 1;
	  tr.mbs -= d_mbs;
	}
      }
#if defined(DEBUG) && DEBUG > 5
      printf ("\t%d\t%d\t%d\t%d\n", 
	      (int)((BITS_PER_CELL*(double)TUNITS)/(double)tr.inv_scr),
	      tr.mbs, d_inv_sbr, d_mbs);
#endif    
    }
    
    if (ok_tr.inv_pcr == -1) {
      printf ("\t%d\t%d", 
	      (int)((BITS_PER_CELL*(double)TUNITS)/(double)tr.inv_scr),
	      tr.mbs);
      printf ("\tFAIL (%d tries)\t", try);
      printf ("%c:%d\n", (inv_clp > 0)?'P':'S', abs(inv_clp));
      exit (1);
    } else {
      printf ("\t%d\t%d", 
	      (int)((BITS_PER_CELL*(double)TUNITS)/(double)ok_tr.inv_scr),
	      ok_tr.mbs);
      printf ("\tOK (%d tries)\n", try);
    }
  }
  exit(0);
}

int not_tried (tr_param *tr, int d_inv_sbr,  int d_mbs)
{
  static int cnt = -1;
  int i, same = 0;
  static tr_param *tr_tbl = NULL;
  static int *sbrt = NULL, *mbst = NULL;

  if (tr == NULL) {
    cnt = 0;
    return 1;
  }
  if (tr_tbl == NULL) {
    if ((tr_tbl = malloc (sizeof(tr_param)*MAX_TRIES)) == NULL ||
	(sbrt = malloc (sizeof(int)*MAX_TRIES)) == NULL ||
	(mbst = malloc (sizeof(int)*MAX_TRIES)) == NULL) {
      perror ("not_tried: ");
      exit (10);
    }
    cnt = 0;
  }

  for (i = 0; i < cnt; i++)
    if (tr_eq (*tr, tr_tbl[i]) && sbrt[i] == d_inv_sbr &&
	mbst[i] == d_mbs) {
      same++;
    }

  sbrt[cnt] = d_inv_sbr;
  mbst[cnt] = d_mbs;
  tr_tbl[cnt] = *tr;
  cnt++;
  return (same <= 2);
}

cell_log read_file (char *fname, int bytes, int inv_lcr, int inv_fcr,
		    int count)
{
  cell_log cl;
  int tstamp;
  int index;			/* current position */
  long long now = 0;
  FILE *fd;

  cl.min = index = 0;

  if ((fd = fopen(fname, "r")) == NULL) {
    perror("fopen");
    exit(2);
  }
  
  cl.max = count * AAL5_CELLS(bytes); 
  if ((cl.table = malloc (sizeof(*cl.table) * cl.max)) == NULL) {
    perror ("malloc");
    exit(2);
  }
  
  while (fscanf (fd, "%d", &tstamp) == 1 && tstamp < 25000000)
    ;
  
  while (fscanf (fd, "%d", &tstamp) == 1 && tstamp < 44994000) {
    int i;
    
    if (now + inv_lcr > (long long)tstamp * 1000) { 
	/* liikaa tavaraa vyllle */
      printf ("Too much traffic at %d (%ld units), now=%ld, cells=%d\n", 
	      tstamp, (long)((now + inv_lcr) - (tstamp*1000)) / 1000,
	      (long)now/1000, index);
      tstamp = (now + inv_lcr) / 1000;
      exit(11);
    }
    for (i = 0; i < AAL5_CELLS (bytes); i++) {
      now = cl.table[index] = ((long long)tstamp) * SCALE + i * inv_fcr;
      if (++index >= cl.max) {
	fclose (fd);		/* jos ei mahdu, vapautetaan 
				 * ja varataan uudestaan */
	free (cl.table);
	return read_file (fname, bytes, inv_lcr, inv_fcr, count * 2);
      }
    }
  }
  cl.max = index - 1;

  return cl;
}
